---
title: "The Fréchet Distance Unleashed: Approximating a Dog with a Frog"
author: Sariel Har-Peled, Benjamin Raichel, Eliot W. Robson
link-citations: true
slideNumber: '`"c/t"`{=html}'
bibliography: fmq.bib
output:
    md_extensions: +grid_tables
initoptions:
    - "width: 20%"
    - "height: 90%"
---

# [frechet.xyz](https://frechet.xyz)

More applied than theory...

--------------------------

# The Fréchet Distance

## Similarity between curves

<video  autoplay loop>
  <source src="https://frechet.xyz/examples/30/f_c_movie.mp4" type="video/mp4">
</video>


Started with [@ag-cfdbt-95]

> - Quadratic time / complicated (use parametric search)

> - Large amount of work on extensions, implementation, etc.



# Frogs taking a walk
## The Discrete Fréchet Distance

![](https://frechet.xyz/examples/01/polygons_sampled.png)


[//]: ############################################################
[//]: ############################################################

-----------

# The problem
## with the Fréchet Distance

<video  autoplay loop>
  <source src="https://frechet.xyz/examples/01/discrete_frechet.mp4" type="video/mp4">
</video>


# The problem II
## with the Fréchet Distance

![](figs/curves_example_3.png){ width="15%" }
![](figs/profile_d_vs_r.png){ width="62.5%" }


[//]: ############################################################
[//]: ############################################################

-------------

# This work

* More nuanced def of Fréchet Distance
* Better + simpler algorithms.
* Implementation (packages)
  + Julia    [https://github.com/sarielhp/FrechetDist.jl](https://github.com/sarielhp/FrechetDist.jl)
  + Python
 [https://github.com/eliotwrobson/FrechetLib](https://github.com/eliotwrobson/FrechetLib)
* Official packages
* New ideas from implementation.


-------------

### The Fréchet distance via free space diagram


![](https://frechet.xyz/examples/10/curves.png){ width="30%" }
![](https://frechet.xyz/examples/10/grid_only.png){ width=65% }


-----------------

### Fréchet distance: Free space diagram II

![](https://frechet.xyz/examples/10/grid_only.png){ width=45% }
![](figs/10_3d.png){ width=50% }


-----------------

## The Fréchet morphing in this case


![](https://frechet.xyz/examples/10/c_diagram.png){ width=50% }
<video width="40%" autoplay loop>
  <source src="https://frechet.xyz/examples/10/f_c_movie.mp4" type="video/mp4">
</video>

-----------------------

[//]: ############################################################
[//]: ############################################################

## First idea: Make moves locally

![](https://frechet.xyz/examples/10/g_diagram.png)

---------------------------------------

### The problem: Bottleneck shortest path

::: columns

:::: column
* Directed graph
* Edges have height
* Compute path from source to target minimize max edge.
  - Min max weight edge.

> * Quadratic time since DAG.

::::

:::: column
![](https://frechet.xyz/examples/10/g_diagram.png){width=100%}
::::

:::

-----------------------


[//]: ############################################################
[//]: ############################################################

### The resulting morphing
#### Vertex-Edge Fréchet (new¿)

:::::::::::::: {.columns}
::: {.column width="50%"}
* Easy to compute.
* Not necessarily monotone!
* But almost monotone!
* Bottleneck shortest path

  For general graphs
  + $O(n \log n + m)$ Dijkstra
  + $O(m \log^* m)$

    [@gt-atbop-88]
:::
::: {.column width="50%"}
![](https://frechet.xyz/examples/10/ve_r_diagram.png)
:::
::::::::::::::

-----------------------

[//]: ############################################################
[//]: ############################################################

## Bottleneck shortest path
### Undirected graphs

* $O(n \log n + m)$: Via MST.

* [@ag-cfdbt-95] alg to compute
  weak Fréchet

  $O(n^2)$.

-----------------------


[//]: ############################################################
[//]: ############################################################

## Retractable Fréchet

### Always keep leash as short as possible

$G$: Directed graph with weights on edges

**Definition**: <text style="color:red;">__Retractable path__</text>
  is the bottleneck shortest path, recursively retractable when
  removing the bottleneck.

-----------

## Retractable definition by example

<div class="r-stack">
  <img
  class="fragment"
  data-fragment-index="0"
  src="figs/steps_1.png"
  width="500"
  />

  <img
  class="fragment"
  data-fragment-index="1"
  src="figs/steps_2.png"
  width="500"
  />

  <img
  class="fragment"
  data-fragment-index="2"
  src="figs/steps_3.png"
  width="500"
  />

  <img
  class="fragment"
  src="figs/steps_4.png"
  data-fragment-index="3"
  width="500"
  />

  <img
  class="fragment"
  data-fragment-index="4"
  src="figs/steps_5.png"
  width="500"
  />
</div>

--------------

### Retractable  Fréchet $\implies$ better morphing

<video width="70%" autoplay loop>
  <source src="https://frechet.xyz/examples/32/discrete_frechet.mp4"
  type="video/mp4">
</video>

<video width="70%" autoplay loop>
  <source
  src="https://frechet.xyz/examples/32/discrete_r_frechet.mp4"
  type="video/mp4">
</video>

----------------

### Computing Retractable Path

* Minor variant of Dijkstra/Prim:
  + From visited vertices: Take cheapest edge
    going out.
  + Prim algorithm for "directed" graph.
  + Stop early!
  - Faster than quadratic for "most" inputs.
  + Algorithm extremely simple (implicit grid).

* Retractable Fréchet
  + Definition due to  [@bbms-lcfm-19]: "Locally correct".
  + Different algorithm


-------------------

Claim: $\forall \cA, \cB$ : $\,\,$
    $\distFr{\cA}{\cB} \geq \distVEFr{\cA}{\cB}$.

![](https://frechet.xyz/examples/10/g_diagram.png)

---------------------

### Non-monotonicity...

<video  autoplay loop>
  <source src="https://frechet.xyz/examples/05/f_ve_r_movie.mp4"
  type="video/mp4">
</video>

![](https://frechet.xyz/examples/05/ve_r_diagram.png){ width="60%" }

---------------------

### Non-monotonicity...
#### Brute force monotonization...

![](https://frechet.xyz/examples/05/ve_r_diagram.png){ width="49%" }
![](https://frechet.xyz/examples/05/ve_r_diagram_m.png){ width="49%" }

Optimal Fréchet $\in[\distVEFrC. \mathrm{mono}(\distVEFrC)]$.

---------------------

### Reduce non-monotonicity by inserting vertices

![](https://frechet.xyz/examples/05/steps/000001.png){ width="35%" }
![](https://frechet.xyz/examples/05/steps/curves_000001.png){ width="59%" }


![](https://frechet.xyz/examples/05/steps/000002.png){ width="35%" }
![](https://frechet.xyz/examples/05/steps/curves_000002.png){ width="59%" }

---------------------

### Reduce non-monotonicity by inserting vertices II

![](https://frechet.xyz/examples/05/steps/000003.png){ width="85%" }

---------------------

### Reduce non-monotonicity by inserting vertices a few iterations later

![](https://frechet.xyz/examples/05/steps/000008.png){ width="85%" }


-----------------------

## Exact $\distFrC$ distance...

- While   $\distVEFrC < \mathrm{mono}(\distVEFrC)$
  + Refine curves
  - Recompute $\distVEFrC(\cA,\cB)$

- isolate the bottleneck...

##

## No such thing as exact algorithms

- Floating point issues.
- Consider $1.0001$-approximation as exact.


---------------------

### The result...

- Algorithm for (exact) Retractable Fréchet

- Retractablity is important...

![](figs/profile_d_vs_r.png){ width="52.5%" }


---------------------

# Approx Fréchet: segment

---------------------

### Comp Fréchet distance segment $\Leftrightarrow$  curve

- $\cA$: Polygonal curve

- $n = |\cA|$

- Exact: $O(n \log n)$.

- Practice: Simple + linear time

-------------------------

![](https://frechet.xyz/examples/33/curves.png){width="39%"}
![](https://frechet.xyz/examples/33/ve_r_diagram.png){width="59%"}

-------------------------

![](https://frechet.xyz/examples/33/curves.png){width="25%"}
![](https://frechet.xyz/examples/33/ve_r_diagram_m.png){width="70%"}

### Lemma
$3$-approx. $O(n)$  time for Fréchet dist between curve and segment.

---------------------

# Simplification: Approx


---------------------

## Greedy Fréchet simplification



![](figs/simp/C_curves_Q.png){width="80%"}

:::: columns

::: {.column width="38%"}
- greedy simplification.

[@ahkww-fdcr-06]
:::
::: {.column width="58%"}
- High quality "slow".

- Idea: Pick vertices greedy fashion...
:::
::::

---------------------

## Approximation via simplification

- $\cA, \cB$: Input curves.

- $d \leftarrow$ silly upper bound on $\distFr{\cA}{\cB}$.

- do
     + $r \leftarrow d/10$
     + $\cA' \leftarrow \mathrm{simplify}(\cA, r )$.
     + $\cB' \leftarrow \mathrm{simplify}(\cB, r )$.
     + $\ell \leftarrow \distFr{\cA'}{\cB'}$
     + Compute morphing $\cA \rightarrow \cA' \rightarrow \cB'
       \rightarrow \cB$
     + If $\frac{\ell +2r}{\ell-2r} \leq 1 + \epsilon$ $\Rightarrow$
       return approx.
     + $d \leftarrow d/4$.

---------------------

# Simplification: Exact


---------------------

### Simplification to make speed great again

- Simplify curves (speedup).

- Do simplification preserving bottleneck.

- $\Rightarrow$ Fast exact algorithm.

- Many details...

----------------------------------------------

### Morphing sensitive simplification I

![](figs/simp/A_original.png){width="48%"}
![](figs/simp/A_P_profile.png){width="48%"}

----------------------------------------------

### Morphing sensitive simplification II

![](figs/simp/A_P_profile.png){width="38%"}
![](figs/simp/A_together.png){width="78%"}


----------------------------------------------

### Morphing sensitive simplification III

![](figs/simp/B_P_profile.png){width="38%"}
![](figs/simp/B_together.png){width="78%"}

----------------------------------------------

### Morphing sensitive simplification IV

![](figs/simp/C_P_profile.png){width="38%"}
![](figs/simp/C_together.png){width="78%"}


----------------------------------------------

# Shticks and tricks
## and questions
### Arising from the implementation

----------------------------------------------


### Low quality simplification, but fast

*Task*: Preprocess $\cA$ ($10,000$s of vertices).

*Query*: Given $\delta > 0$ output curve $\cB$ s.t.
$\distFr{\cA}{\cB} \leq \delta$.

*Running time*: $O( |V(\cB)|)$.

---------------------

## Output sensitive simplification...

::: {.r-stack}

![](figs/simp_1.png){ width="450" }

![](figs/simp_2.png){.fragment width="450" }

![](figs/simp_3.png){.fragment width="450" }

![](figs/simp_4.png){.fragment width="450" }

![](figs/simp_5.png){.fragment width="450" }

:::

Given $\alpha$: Extract *all* vertices label $>\alpha$.

Running time: Output sensitive.


-----------------------------------

## Simplifications hierarchy/caching

- repeats many Fréchet distance queries on same curves.

- Cache simplifications curves (a hierarchy).

- use hierarchy to extract simplification quickly.

- Precompute the simplification hierarchy.

- Good idea if persistent data/task...

-----------------------------------

# Implementations

-----------------------------------

## Julia (Programming language)

- Python without the stupidity.

- As much typed as you want.

- built-in package manager

- Many libraries.


- Source code of everything available.

- Compiled on execution.

- Designed for multi-threaded/parallel and distributed computing
- Fun.

-----------------------------------

## ACM SIGSPATIAL Cup 2017 

> The Fréchet distance is defined as the minimal length of a leash
> connecting a dog on one trajectory with its owner on a second
> trajectory, both never moving backwards. For this challenge, the
> "true" Fréchet distance shall be used, though teams are free to
> exploit the existing approximations including the Discrete Fréchet
> distance as they wish. The challenge for this year consists of finding
> and combining spatiotemporal   indexing schemes with efficient
> algorithms for Range Queries in very large databases of
> trajectories...

-----------------------------------

# C++

[@bkn-wdfpa-21]: C++ implementation.


[//]: ############################################################
[//]: ############################################################

# Implementation

## Julia+Python package.


# Experiments

[@bkn-wdfpa-21]: C++ implementation.

![](http://sarielhp.org/misc/blog/24/09/23/julia_cpp.png)

--------------------------------------

# Inputs

![](http://sarielhp.org/misc/blog/24/09/23/julia_results_2.gif)

--------------------------------------

# Julia results

![](http://sarielhp.org/misc/blog/24/09/23/julia_results.gif)

-------------------------------------

# Python results

![](http://sarielhp.org/misc/blog/24/09/23/python_results.gif)

-------------------------------------

# Future research

-------------------------------------

## Future research

- Handling noise/outliers.

- Consensus curve?



# Bibliography

```{=html}
<style>
body { min-width: 100% !important; }
</style>
```
